
// ================================================================ \\
// NamedPipeClient (Source Side)
#define GENERIC_WRITE                  0x40000000
#define OPEN_EXISTING                  3
#define INVALID_HANDLE_VALUE           -1
#import "kernel32.dll"
   int CreateFileW(string Filename, int AccessMode, int ShareMode, int PassAsZero, int CreationMode, int FlagsAndAttributes, int AlsoPassAsZero);
   int CloseHandle(int fileHandle);
   //int WriteFile(int FileHandle, string Buffer, int BufferLength, int & BytesWritten[], int PassAsZero);
   int WriteFile(int FileHandle, uchar & Buffer[], int BufferLength, int & BytesWritten[], int PassAsZero);
#import
#define PIPE_BASE_NAME     "\\\\.\\pipe\\mt4-"     // MUST tally with server code
// Sends a message to a server pipe
bool SendPipeMessage( string ToPipe, string Message ){
   bool bReturnvalue = false;
   string strPipeName = StringConcatenate(PIPE_BASE_NAME,ToPipe);
   int PipeHandle = CreateFileA(strPipeName, GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
   if (PipeHandle == INVALID_HANDLE_VALUE) {
      bReturnvalue = false;
   }
   else{
      int BytesWritten[1] = {0};
      WriteFile(PipeHandle, Message, StringLen(Message), BytesWritten, 0);
      if (BytesWritten[0] > 0) {
         bReturnvalue = true;
      } else {
         bReturnvalue = false;
      }
      CloseHandle(PipeHandle);
   }
   return(bReturnvalue);
}

// ================================================================ \\
// NamedPipeServer (Client Side)
#define DEFAULT_MAX_PIPES  20
#define GENERIC_READ                   0x80000000
#define GENERIC_WRITE                  0x40000000
#define PIPE_ACCESS_DUPLEX             3
#define PIPE_UNLIMITED_INSTANCES       255
#define PIPE_NOWAIT                    1
#define PIPE_TYPE_MESSAGE              4
#define PIPE_READMODE_MESSAGE          2   
#define PIPE_WAIT                      0
#import "kernel32.dll"
   int CreateNamedPipeW(string pipeName,int openMode,int pipeMode,int maxInstances,int outBufferSize,int inBufferSize,int defaultTimeOut,int security);
   int PeekNamedPipe(int PipeHandle, int PassAsZero, int PassAsZero2, int PassAsZero3, int & BytesAvailable[], int PassAsZero4);
   //int CreateFileA(string Filename, int AccessMode, int ShareMode, int PassAsZero, int CreationMode, int FlagsAndAttributes, int AlsoPassAsZero);
   int CloseHandle(int fileHandle);
   //int ReadFile(int FileHandle, int BufferPtr, int BufferLength, int & BytesRead[], int PassAsZero);
   int ReadFile(int FileHandle, uchar & BufferPtr[], int BufferLength, int & BytesRead[], int PassAsZero);
   int MulDiv(string X, int N1, int N2);
#import
int glbPipeCount = DEFAULT_MAX_PIPES;
int glbPipe[DEFAULT_MAX_PIPES];
string glbPipeName;
// Starts the pipe server by creating n instances of the pipe
void CreatePipeServer(string PipeName, int UsePipeInstances = DEFAULT_MAX_PIPES){
   // Store the number of pipe instances to use and resize the array accordinging
   glbPipeCount = UsePipeInstances;
   ArrayResize(glbPipe, glbPipeCount);
   // Store the name to use for the pipe instances 
   glbPipeName = PipeName;
   // Create the pipe instances
   for (int i = 0; i < glbPipeCount; i++) {
      glbPipe[i] = CreatePipeInstance();
   }
   return;
}
void DestroyPipeServer(){
   for (int i = 0; i < glbPipeCount; i++) {
      CloseHandle(glbPipe[i]);
   }
   return;
}
int CheckForPipeMessages(string & arrMessages[]){
   int MessagesFound = 0;
   for (int i = 0; i < glbPipeCount; i++) {
      string strMsg = CheckPipe(i);
      if (strMsg != "") {
         MessagesFound++;
         ArrayResize(arrMessages, MessagesFound);
         arrMessages[MessagesFound - 1] = strMsg;
      }
   }
   return (MessagesFound);
}
string CheckPipe(int PipeIndex){
   string strReturnValue = "";
   // See if there's data available on the pipe   
   int BytesAvailable[1] = {0};
   int res = PeekNamedPipe(glbPipe[PipeIndex], 0, 0, 0, BytesAvailable, 0);
   if (res != 0) {
      // PeekNamedPipe() succeeded
      // Is there data?
      if (BytesAvailable[0] != 0) {
         // Keep reading until either we have all the data, or an error occurs         
         int TotalBytesRead = 0;
         while (TotalBytesRead < BytesAvailable[0]) {            
            // Allocate a 200-byte buffer
            string ReadBuffer = "01234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789";
            int BufferLength = StringLen(ReadBuffer);
            // Read up to the maximum buffer size from the pipe. (The use of MulDiv() is an old trick from VB6 days for 
            // getting the address in memory of a string variable.)
            int BytesRead[1] = {0};
            ReadFile(glbPipe[PipeIndex], MulDiv(ReadBuffer, 1, 1), BufferLength, BytesRead, 0);
            // Did we get any data from the read?
            if (BytesRead[0] > 0) {
               // Yes, got some data. Add it to the total message which is passed back
               strReturnValue = StringConcatenate(strReturnValue, StringSubstr(ReadBuffer, 0, BytesRead[0]));   
               TotalBytesRead += BytesRead[0];
            } else {
               // No, the read failed. Stop reading, and pass back an empty string 
               strReturnValue = "";
               TotalBytesRead = 999999;
            }
         }
         // Destroy and recreate the pipe instance
         CloseHandle(glbPipe[PipeIndex]);
         glbPipe[PipeIndex] = CreatePipeInstance();
      } else {
         // No data available on pipe 
      }
   } else {
      // PeekNamedPipe() failed
   }
   return (strReturnValue);
}
// Service function which creates a pipe instance
int CreatePipeInstance(){
   string strPipeName = StringConcatenate(PIPE_BASE_NAME , glbPipeName);
   return (CreateNamedPipeA(strPipeName, GENERIC_READ | PIPE_ACCESS_DUPLEX, PIPE_TYPE_MESSAGE | PIPE_READMODE_MESSAGE | PIPE_NOWAIT, PIPE_UNLIMITED_INSTANCES, 1000, 1000, 0, NULL));
}

